package Interp;

import java.util.LinkedList;
import java.io.*;

import Parse.Token;
import Absyn.*;
import NLObject.*;
import Sculpture.*;


public class Interp
{

    public static boolean FOR_C_DRIVER = false;
    static boolean C_DRIVER_FIRST = true;
    static StringBuffer patDump;
    static StringBuffer durDump;
    static int npats;

    SymTab symTab;
    Sculpture sculpt;

    public Interp(LinkedList stmts, Sculpture sculpt)
    {
        symTab = new SymTab();
        this.sculpt = sculpt;

        // put constants in symTab
        symTab.bind("SIZE", new Int(Pat.SIZE));

        if(FOR_C_DRIVER)
        {
            patDump = new StringBuffer();
            durDump = new StringBuffer();
            npats = 0;
        }

        for(int i=0; i<stmts.size(); i++)
        {
            Stmt s = (Stmt)stmts.get(i);
            interpStmt(s);
        }

        if(FOR_C_DRIVER)
        {
            System.out.println("#define SIZE " + Pat.SIZE);
            System.out.println("#define NPATS " + npats);

            System.out.println("\n");

            System.out.println("char pat[" + npats + "][" + Pat.SIZE + "] = {");
            System.out.print(patDump);
            System.out.println("\n};\n");

            System.out.println("int dur[" + npats + "] = {");
            System.out.print(durDump);
            System.out.println("\n};\n");
        }
    }

    public NLObject interpStmt(Stmt s)
    {
        // short circuit erroneous interpretation
        if(errorFlag) return null;

        if(s instanceof Line)
            return interpLine((Line)s);

        if(s instanceof Block)
            return interpBlock((Block)s);

        if(s instanceof If)
            return interpIf((If)s);

        if(s instanceof While)
            return interpWhile((While)s);

        if(s instanceof For)
            return interpFor((For)s);

        if(s instanceof Func)
            return interpFunc((Func)s);

        error("Unrecognized statement: " + s);
        return null;
    }

    public NLObject interpFunc(Func f)
    {
        symTab.bind(f.id.image, f);
        return null;
    }

    public NLObject interpFor(For f)
    {
        String id = f.id.image;

        NLObject lObj = interpExpr(f.list);

        if(errorFlag) return null;
        if(! (lObj instanceof List))
        {
            error("At line: " + f.id.beginLine);
            error("<expr> in 'for( <id> in <expr> )' was " + lObj.getType());
            error("must be list instead.");
            return null;
        }

        List list = (List)lObj;

        for(int i=0; i<list.list.size(); i++)
        {
            symTab.bind(id, (NLObject)list.list.get(i));

            NLObject res = interpStmt(f.body);

            if(errorFlag) return null;

            if(res != null && 
               !(res instanceof Break) && 
               !(res instanceof Continue))
                return res;

            if(res instanceof Break)
                break;
        }

        return null;
    }

    public NLObject interpWhile(While w)
    {
        NLObject cond = interpExpr(w.cond);

        if( ! (cond instanceof Bool))
            error("While condition must be a boolean value.");

        if(errorFlag && out.indexOf("Near line") > -1)
        {
            error("Near line " + w.line + ".");
            return null;
        }

        while( ((Bool)cond).bool && !errorFlag )
        {
            NLObject res = interpStmt(w.body);

            if(res != null && 
               !(res instanceof Break) && 
               !(res instanceof Continue))
                return res;

            if(res instanceof Break)
                break;

            // update condition
            cond = interpExpr(w.cond);

            if( ! (cond instanceof Bool))
                error("While condition must be a boolean value.");

            if(errorFlag && out.indexOf("Near line") > -1)
            {
                error("Near line " + w.line + ".");
                return null;
            }
        }

        return null;
    }

    public NLObject interpIf(If i)
    {
        NLObject cond = interpExpr(i.cond);

        if( ! (cond instanceof Bool))
            error("If condition must be a boolean value.");

        if(errorFlag && out.indexOf("Near line") > -1)
        {
            error("Near line " + i.line + ".");
            return null;
        }

        if( ((Bool)cond).bool )
            return interpStmt(i.thenS);

        else if(i.elseS != null)
            return interpStmt(i.elseS);

        return null;
    }

    public NLObject interpBlock(Block b)
    {
        for(int i=0; i<b.stmts.size(); i++)
        {
            NLObject nlo = interpStmt((Stmt)b.stmts.get(i));

            // kill erroneous interpretation
            if(errorFlag) return null;

            if(nlo != null) return nlo;
        }

        return null;
    }

    public NLObject interpLine(Line l)
    {
        // provide hints with error
        if(errorFlag && out.indexOf("Near line") > -1)
        {
            error("Near line " + l.lineNumber + ".");
            return null;
        }

        if(l instanceof BindLine)
            return interpBindLine((BindLine)l);

        if(l instanceof PrintLine)
            return interpPrintLine((PrintLine)l);

        if(l instanceof SculptLine)
            return interpSculptLine((SculptLine)l);

        if(l instanceof BreakLine)
            return interpBreakLine((BreakLine)l);

        if(l instanceof ContLine)
            return interpContLine((ContLine)l);

        if(l instanceof CallLine)
            return interpCallLine((CallLine)l);

        if(l instanceof RetLine)
            return interpRetLine((RetLine)l);

        error("Unrecognized line: " + l);
        return null;
    }

    public NLObject interpRetLine(RetLine rl)
    {
        if(rl.ret != null)
            return interpExpr(rl.ret);
        return new NLVoid();
    }

    public NLObject interpCallLine(CallLine cl)
    {
        return interpCall(cl.call);
    }

    public NLObject interpContLine(ContLine cl)
    {
        return new Continue();
    }

    public NLObject interpBreakLine(BreakLine bl)
    {
        return new Break();
    }

    public NLObject interpBindLine(BindLine bl)
    {
        interpBind(bl.bind);
        return null;
    }

    public NLObject interpSculptLine(SculptLine sl)
    {
        NLObject pObj = interpExpr(sl.pat);
        NLObject dObj = interpExpr(sl.dur);

        if(pObj instanceof Pat && dObj instanceof Int)
        {
            Pat pat = (Pat)pObj;
            int dur = ((Int)dObj).i;

            if(FOR_C_DRIVER)
            {
                if(C_DRIVER_FIRST)
                {
                    patDump.append( "\"" + pat.toString() + "\"" );
                    durDump.append( "" + dur );
                    C_DRIVER_FIRST = false;
                }
                else
                {
                    patDump.append( ",\n\"" + pat.toString() + "\"" );
                    durDump.append( ",\n" + dur );
                }

                npats++;
            }
            else
            {
                sculpt.sculpt(pat);

                try
                {
                    Thread.sleep(10 * dur);
                }
                catch(Exception e)
                { }

            }

            return null;
        }

        error("Incorrect types for sculpt.");
        error("Should be: sculpt( <pattern> , <integer> )");
        return null;
    }

    public NLObject interpPrintLine(PrintLine pl)
    {
        for(int i=0; i<pl.exprs.size(); i++)
        {
            Expr e = (Expr)pl.exprs.get(i);
            NLObject nlo = interpExpr(e);
            if(nlo == null)
            {
                out.append(" (null) ");
            }
            else
            {
                out.append(nlo.toString());
            }
        }
        out.append("\n");
        return null;
    }


    public NLObject interpExpr(Expr e)
    {
        if(e instanceof Bind)
            return interpBind((Bind)e);

        if(e instanceof Literal)
            return interpLiteral((Literal)e);

        if(e instanceof ListCon)
            return interpListCon((ListCon)e);

        if(e instanceof BinOp)
            return interpBinOp((BinOp)e);

        if(e instanceof UnOp)
            return interpUnOp((UnOp)e);

        if(e instanceof IdExpr)
            return interpIdExpr((IdExpr)e);

        if(e instanceof Call)
            return interpCall((Call)e);

        if(e instanceof ListSel)
            return interpListSel((ListSel)e);

        error("Unrecognized expression: " + e);
        return null;
    }

    public NLObject interpUnOp(UnOp u)
    {
        NLObject opand = interpExpr(u.expr);
        return opand.op(u.op.image, new NLVoid());
    }

    public NLObject interpListSel(ListSel l)
    {
        NLObject list = interpExpr(l.list);

        if(! (list instanceof List))
        {
            error("Trying to select element of non-list.");
            return null;
        }

        NLObject ind = interpExpr(l.ind);

        if(! (ind instanceof Int))
        {
            error("Trying to select element of non-list.");
            return null;
        }

        return ((List)list).getElt(((Int)ind).i);
    }

    public NLObject interpCall(Call c)
    {
        NLObject fObj = interpExpr(c.func);

        if(errorFlag) return null;

        if(! (fObj instanceof Func))
        {
            error("Call to non function: " + fObj.toString() +
                  " (" + fObj.getType() + ")");
            return null;
        }

        Func func = (Func)fObj;

        if(func.params.size() != c.args.size())
        {
            error("Wrong number of arguments to: " + func.toString());
            return null;
        }

        LinkedList args = new LinkedList();

        symTab.startScope();

        // evaluate args and load them up into symTab
        for(int i=0; i<c.args.size(); i++)
        {
            NLObject arg = interpExpr((Expr)c.args.get(i));
            String id = ((Token)func.params.get(i)).image;
            symTab.bind(id, arg);

            // save args in case builtin functions need them
            args.add(arg);
        }

        NLObject ret;

        // handle builtins
        if(func.id.image.equals("length"))
        {
            ret = length((NLObject)(args.get(0)));
        }
        else // evaluate body
        {
            ret = interpStmt(func.body);
        }

        symTab.endScope();

        return ret;
    }

    public NLObject length(NLObject list)
    {
        if(! (list instanceof List))
        {
            error("Cannot take the length of a non-list.");
            return null;
        }

        return new Int(((List)list).list.size());
    }

    public NLObject interpListCon(ListCon lc)
    {
        LinkedList nlos = new LinkedList();

        for(int i=0; i<lc.exprs.size(); i++)
        {
            NLObject n = interpExpr((Expr)lc.exprs.get(i));

            if(errorFlag) return null;

            nlos.add(n);
        }

        return new List(nlos);
    }

    public NLObject interpBind(Bind b)
    {
        String id = b.id.image;
        NLObject val = interpExpr(b.val);

        if(val == null || val instanceof NLVoid)
        {
            error("Cannot bind to non-value.");
            return null;
        }

        symTab.bind(id, val);
        return val;
    }

    public NLObject interpIdExpr(IdExpr ie)
    {
        String id = ie.id.image;
        NLObject val = symTab.lookup(id);

        if(val == null)
        {
            error("Id '" + id + "' is not bound to any value.");
            return null;
        }

        return val;
    }

    public NLObject interpLiteral(Literal l)
    {
        return l.nlo;
    }

    public NLObject interpBinOp(BinOp bo)
    {
        String   op    = bo.op.image;
        NLObject left  = interpExpr(bo.left); 
        NLObject right = interpExpr(bo.right); 

        if(left == null || right == null)
        {
            error("Bad operands to '" + op + "'.");
            return null;
        }

        return left.op(op, right);
    }


    /* ERROR HANDLING */
    public static boolean errorFlag = false;
    public static StringBuffer out  = new StringBuffer();

    public static void error(String msg)
    {
        errorFlag = true;
        out.append("Error: " + msg + "\n");
    }

    public static void warn(String msg)
    {
        out.append("Warning: " + msg + "\n");
    }
}
